# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: metagriffin <mg.github@uberdev.org>
# date: 2013/12/10
# copy: (C) Copyright 2013-EOT metagriffin -- see LICENSE.txt
#------------------------------------------------------------------------------
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#------------------------------------------------------------------------------

class Source(object):
  '''
  Source is a helper object with input streams to:

  a. Normalize line endings to "\n"
  b. Provide line-oriented stream peeking.
  '''
  def __init__(self, stream):
    self.stream  = stream
    self.cache   = []
    self.curline = 1
  def readline(self):
    self.curline += 1
    if self.cache:
      return self.cache.pop()
    ret = self.stream.readline()
    if not ret:
      return ret
    if ret.endswith('\r'):
      ret = ret[:-1] + '\n'
    elif ret.endswith('\r\n'):
      ret = ret[:-2] + '\n'
    return ret
  def readlines(self, count):
    return [self.readline() for idx in range(count)]
  def pushline(self, line):
    self.curline -= 1
    if line:
      self.cache.append(line)
  def pushlines(self, lines):
    for line in reversed(lines):
      self.pushline(line)
  def peekline(self):
    line = self.readline()
    self.pushline(line)
    return line
  def peeklines(self, count):
    ret = self.readlines(count)
    self.pushlines(ret)
    return ret

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: metagriffin <mg.github@uberdev.org>
# date: 2013/11/27
# copy: (C) Copyright 2013-EOT metagriffin -- see LICENSE.txt
#------------------------------------------------------------------------------
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#------------------------------------------------------------------------------

import os, unittest, asset, re
from aadict import aadict

import parsedifflib
from parsedifflib import api

heredir = os.path.abspath(os.path.dirname(__file__))
datadir = os.path.join(heredir, 'data')

#------------------------------------------------------------------------------
def wrap(data, start, end):
  data = (end + '\n' + start).join(data.split('\n'))
  return start + data + end
def colorize(data, style='svnlook'):
  spec = aadict(
    add    = lambda data: wrap(data, '[1;34m', '[0;0m'),
    delete = lambda data: wrap(data, '[1;31m', '[0;0m'),
    note   = lambda data: wrap(data, '[1;35m', '[0;0m'),
    plain  = lambda data: wrap(data, '[1;37m', '[0;0m'),
    meta   = lambda data: wrap(data, '[1;32m', '[0;0m'),
    )
  if style != 'svnlook':
    raise ValueError('unsupported colorization style: ' + style)
  ret = []
  ent = 0

  for etype, obj in parsedifflib.parse_svnlook(data):

    if etype in (
        parsedifflib.Event.PATCH_START,
        parsedifflib.Event.PATCH_END,
        ):
      continue

    elif etype == parsedifflib.Event.ENTRY_START:
      ent += 1
      if ent > 1:
        ret.append(spec.plain(''))
      if obj.type == parsedifflib.Entry.TYPE_PROPERTY:
        ret.append(spec.plain(''))
      ret.append(spec.meta(obj.comment))
      if obj.type == parsedifflib.Entry.TYPE_PROPERTY:
        ret.append(spec.meta('_' * 67))
      else:
        ret.append(spec.meta('=' * 67))
      if obj.source:
        ret.append(spec.delete('--- {}\t{}'.format(obj.source, obj.srcsig or '')))
      if obj.target:
        ret.append(spec.add('+++ {}\t{}'.format(obj.target, obj.tgtsig or '')))

    elif etype == parsedifflib.Event.ENTRY_END:
      if obj.type == parsedifflib.Entry.TYPE_PROPERTY:
        ret.append(spec.plain(''))
        # todo: HACK ALERT! i don't know why this is necessary. it appears
        # that sometimes svnlook does not always insert this blank line...
        ent = 0

    elif etype == parsedifflib.Event.LINE_LOC:
      ret.append(spec.note(obj.line))

    elif etype == parsedifflib.Event.LINE_NOTE:
      ret.append(spec.note(obj.line))

    elif etype == parsedifflib.Event.LINE_SAME:
      ret.append(spec.plain(' ' + obj.line))

    elif etype == parsedifflib.Event.LINE_DELETE:
      ret.append(spec.delete('-' + obj.line))

    elif etype == parsedifflib.Event.LINE_ADD:
      ret.append(spec.add('+' + obj.line))

    elif etype == parsedifflib.Event.PROPENTRY:
      ret.append(spec.meta(obj.head))
      if obj.old:
        ret.append(spec.delete('   - ' + obj.old))
      if obj.new:
        ret.append(spec.add('   + ' + obj.new))

    else:
      raise ValueError('unexpected event type: ' + repr(etype))


  return '\n'.join(ret) + '\n'

#------------------------------------------------------------------------------
class TestParseDiffLib(unittest.TestCase):

  maxDiff = None

  nietzsche_diff = '''\
Modified: Quotes/FriedrichNietzsche-TheWay.txt
===================================================================
--- Quotes/FriedrichNietzsche-TheWay.txt	1844-10-15 00:00:00 UTC (rev 1)
+++ Quotes/FriedrichNietzsche-TheWay.txt	1900-08-25 00:00:00 UTC (rev 2)
@@ -1,7 +1,7 @@
 You have your way. I have my
 way. As for the right way, the
 correct way, and the only way,
-it does not exist.
+it's definitely not your way ;).
 
   -- Friedrich Nietzsche
      1844 - 1900
'''

  propEntry_diff = '''\

Property changes on: property/file1.txt
___________________________________________________________________
Modified: test:property1
   - one line

(blank line above and below too)


   + one line
(blank line above removed)
(blank line below)



'''

  #----------------------------------------------------------------------------
  def test_raw(self):
    patch = api.Patch()
    ent1 = api.Entry(
      api.Entry.TYPE_CONTENT,
      'app/bin/runner', 'app/bin/runner',
      '2013-05-31 16:43:09 UTC (rev 2958)',
      '2013-05-31 17:54:22 UTC (rev 2959)',
      'Deleted: app/bin/runner')
    ent2 = api.Entry(
      api.Entry.TYPE_CONTENT,
      'app/setup.py', 'app/setup.py',
      '2013-05-31 16:43:09 UTC (rev 2958)',
      '2013-05-31 17:54:22 UTC (rev 2959)',
      'Modified: app/setup.py')
    ent3 = api.Entry(
      api.Entry.TYPE_CONTENT,
      'app/lib/python/MANIFEST', 'app/lib/python/MANIFEST',
      '2013-05-31 16:43:09 UTC (rev 2958)',
      '2013-05-31 17:54:22 UTC (rev 2959)',
      'Modified: app/lib/python/MANIFEST')
    ent4 = api.Entry(
      api.Entry.TYPE_CONTENT,
      comment = 'Added: app/lib/python/my_app-1.2.3.tar.gz')
    ent5 = api.Entry(
      api.Entry.TYPE_PROPERTY,
      comment = 'Property changes on: app/lib/python/my_app-1.2.3.tar.gz')
    chk = [
      (api.Event.PATCH_START, patch),
      (api.Event.ENTRY_START, ent1),
      (api.Event.LINE_LOC, api.Line('@@ -1,4 +0,0 @@')),
      (api.Event.LINE_DELETE, api.Line('#!/usr/bin/env python')),
      (api.Event.LINE_DELETE, api.Line('import sys')),
      (api.Event.LINE_DELETE, api.Line('from app.cli import main')),
      (api.Event.LINE_DELETE, api.Line('sys.exit(main())')),
      (api.Event.ENTRY_END, ent1),
      (api.Event.ENTRY_START, ent2),
      (api.Event.LINE_LOC, api.Line('@@ -80,7 +80,7 @@')),
      (api.Event.LINE_SAME, api.Line('  \'Shapely                  == 1.2.17\',')),
      (api.Event.LINE_SAME, api.Line('')),
      (api.Event.LINE_SAME, api.Line('  # message queue')),
      (api.Event.LINE_DELETE, api.Line('  # \'my-app                   == x.x.x\',')),
      (api.Event.LINE_ADD, api.Line('  \'my-app                   == 1.2.3\',')),
      (api.Event.LINE_SAME, api.Line('  \'APScheduler              == 2.1.0\',')),
      (api.Event.LINE_SAME, api.Line('  \'kombu                    == 2.5.10\',')),
      (api.Event.LINE_SAME, api.Line('  \'amqp                     == 1.0.11\',')),
      (api.Event.ENTRY_END, ent2),
      (api.Event.ENTRY_START, ent3),
      (api.Event.LINE_LOC, api.Line('@@ -68,4 +68,5 @@')),
      (api.Event.LINE_SAME, api.Line('https://pypi.python.org/packages/source/G/GeoAlchemy2/GeoAlchemy2-0.2.tar.gz#md5=14d73d09cdc47e3ed92e804883e61218')),
      (api.Event.LINE_SAME, api.Line('https://pypi.python.org/packages/source/t/transaction/transaction-1.4.1.zip#md5=8db2680bc0f999219861a67b8f335a88')),
      (api.Event.LINE_SAME, api.Line('https://pypi.python.org/packages/source/p/psycopg2/psycopg2-2.5.tar.gz#md5=facd82faa067e99b80146a0ee2f842f6')),
      (api.Event.LINE_DELETE, api.Line('https://pypi.python.org/packages/source/p/pyramid/pyramid-1.4.1.tar.gz#md5=044d42f609d567d7db2948a03fffcf7c')),
      (api.Event.LINE_NOTE, api.Line('\ No newline at end of file')),
      (api.Event.LINE_ADD, api.Line('https://pypi.python.org/packages/source/p/pyramid/pyramid-1.4.1.tar.gz#md5=044d42f609d567d7db2948a03fffcf7c')),
      (api.Event.LINE_ADD, api.Line('https://pypi.python.org/packages/source/m/my_app/my_app-1.2.3.tar.gz')),
      (api.Event.LINE_NOTE, api.Line('\ No newline at end of file')),
      (api.Event.ENTRY_END, ent3),
      (api.Event.ENTRY_START, ent4),
      (api.Event.LINE_NOTE, api.Line('(Binary files differ)')),
      (api.Event.ENTRY_END, ent4),
      (api.Event.ENTRY_START, ent5),
      (api.Event.PROPENTRY, api.PropertyEntry('Added: svn:mime-type', None, 'application/octet-stream')),
      (api.Event.ENTRY_END, ent5),
      (api.Event.PATCH_END, patch),
      ]

    data = asset.load('parsedifflib:data/svnlook.diff').read()
    gen = parsedifflib.parse_svnlook(data)
    chk = [(type, repr(obj)) for type, obj in chk]
    out = [(type, repr(obj)) for type, obj in gen]
    self.assertEqual(out, chk)

  #----------------------------------------------------------------------------
  def test_lineNumbers(self):
    src   = self.nietzsche_diff
    patch = api.Patch()
    entry = api.Entry(
      api.Entry.TYPE_CONTENT,
      'Quotes/FriedrichNietzsche-TheWay.txt',
      'Quotes/FriedrichNietzsche-TheWay.txt',
      '1844-10-15 00:00:00 UTC (rev 1)',
      '1900-08-25 00:00:00 UTC (rev 2)',
      'Modified: Quotes/FriedrichNietzsche-TheWay.txt')
    chk = [
      (api.Event.PATCH_START, patch),
      (api.Event.ENTRY_START, entry),
      (api.Event.LINE_LOC, api.Line('@@ -1,7 +1,7 @@')),
      (api.Event.LINE_SAME, api.Line('You have your way. I have my', oldnum=1, newnum=1)),
      (api.Event.LINE_SAME, api.Line('way. As for the right way, the', oldnum=2, newnum=2)),
      (api.Event.LINE_SAME, api.Line('correct way, and the only way,', oldnum=3, newnum=3)),
      (api.Event.LINE_DELETE, api.Line('it does not exist.', oldnum=4)),
      (api.Event.LINE_ADD, api.Line('it\'s definitely not your way ;).', newnum=4)),
      (api.Event.LINE_SAME, api.Line('', oldnum=5, newnum=5)),
      (api.Event.LINE_SAME, api.Line('  -- Friedrich Nietzsche', oldnum=6, newnum=6)),
      (api.Event.LINE_SAME, api.Line('     1844 - 1900', oldnum=7, newnum=7)),
      (api.Event.ENTRY_END, entry),
      (api.Event.PATCH_END, patch),
      ]
    gen = parsedifflib.parse_svnlook(src, {'lineNumbers': True})
    chk = [(type, repr(obj)) for type, obj in chk]
    out = [(type, repr(obj)) for type, obj in gen]
    self.assertEqual(out, chk)

  #----------------------------------------------------------------------------
  def test_lineNumbers_singleLine(self):
    src = '''\
Modified: content/textfile.txt
===================================================================
--- content/textfile.txt	2011-04-29 02:40:55 UTC (rev 2)
+++ content/textfile.txt	2011-04-29 02:57:23 UTC (rev 3)
@@ -1 +1 @@
-this is a sample textfile.txt.
+this is a sample textfile.txt with changes.
'''
    patch = api.Patch()
    entry = api.Entry(
      api.Entry.TYPE_CONTENT,
      'content/textfile.txt', 'content/textfile.txt',
      '2011-04-29 02:40:55 UTC (rev 2)',
      '2011-04-29 02:57:23 UTC (rev 3)',
      'Modified: content/textfile.txt')
    chk = [
      (api.Event.PATCH_START, patch),
      (api.Event.ENTRY_START, entry),
      (api.Event.LINE_LOC, api.Line('@@ -1 +1 @@')),
      (api.Event.LINE_DELETE, api.Line('this is a sample textfile.txt.', oldnum=1)),
      (api.Event.LINE_ADD, api.Line('this is a sample textfile.txt with changes.', newnum=1)),
      (api.Event.ENTRY_END, entry),
      (api.Event.PATCH_END, patch),
      ]
    gen = parsedifflib.parse_svnlook(src, {'lineNumbers': True})
    chk = [(type, repr(obj)) for type, obj in chk]
    out = [(type, repr(obj)) for type, obj in gen]
    self.assertEqual(out, chk)

  #----------------------------------------------------------------------------
  def test_intraLineDiff(self):
    src   = self.nietzsche_diff
    patch = api.Patch()
    entry = api.Entry(
      api.Entry.TYPE_CONTENT,
      'Quotes/FriedrichNietzsche-TheWay.txt',
      'Quotes/FriedrichNietzsche-TheWay.txt',
      '1844-10-15 00:00:00 UTC (rev 1)',
      '1900-08-25 00:00:00 UTC (rev 2)',
      'Modified: Quotes/FriedrichNietzsche-TheWay.txt')
    chk = [
      (api.Event.PATCH_START, patch),
      (api.Event.ENTRY_START, entry),
      (api.Event.LINE_LOC, api.Line('@@ -1,7 +1,7 @@')),
      (api.Event.LINE_SAME, api.Line('You have your way. I have my', oldnum=1, newnum=1)),
      (api.Event.LINE_SAME, api.Line('way. As for the right way, the', oldnum=2, newnum=2)),
      (api.Event.LINE_SAME, api.Line('correct way, and the only way,', oldnum=3, newnum=3)),
      (api.Event.LINE_DELETE, api.Line('it does not exist.', oldnum=4, segments=[
        (api.Event.SEGMENT_SAME, 'it d'),
        (api.Event.SEGMENT_DELETE, 'o'),
        (api.Event.SEGMENT_SAME, 'e'),
        (api.Event.SEGMENT_DELETE, 's'),
        (api.Event.SEGMENT_SAME, ' not '),
        (api.Event.SEGMENT_DELETE, 'exist'),
        (api.Event.SEGMENT_SAME, '.'),
        ])),
      (api.Event.LINE_ADD, api.Line('it\'s definitely not your way ;).', newnum=4, segments=[
        (api.Event.SEGMENT_SAME, 'it'),
        (api.Event.SEGMENT_ADD, '\'s'),
        (api.Event.SEGMENT_SAME, ' de'),
        (api.Event.SEGMENT_ADD, 'finitely'),
        (api.Event.SEGMENT_SAME, ' not '),
        (api.Event.SEGMENT_ADD, 'your way ;)'),
        (api.Event.SEGMENT_SAME, '.'),
        ])),
      (api.Event.LINE_SAME, api.Line('', oldnum=5, newnum=5)),
      (api.Event.LINE_SAME, api.Line('  -- Friedrich Nietzsche', oldnum=6, newnum=6)),
      (api.Event.LINE_SAME, api.Line('     1844 - 1900', oldnum=7, newnum=7)),
      (api.Event.ENTRY_END, entry),
      (api.Event.PATCH_END, patch),
      ]
    gen = parsedifflib.parse_svnlook(src, {'lineNumbers': True, 'intraLineDiff': True})
    chk = [(type, repr(obj)) for type, obj in chk]
    out = [(type, repr(obj)) for type, obj in gen]
    self.assertEqual(out, chk)

  #----------------------------------------------------------------------------
  def test_intraLineDiff_singleLine(self):
    src = '''\
Modified: content/textfile.txt
===================================================================
--- content/textfile.txt	2011-04-29 02:40:55 UTC (rev 2)
+++ content/textfile.txt	2011-04-29 02:57:23 UTC (rev 3)
@@ -1 +1 @@
-this is a sample textfile.txt.
+this is a sample textfile.txt with changes.
'''
    patch = api.Patch()
    entry = api.Entry(
      api.Entry.TYPE_CONTENT,
      'content/textfile.txt', 'content/textfile.txt',
      '2011-04-29 02:40:55 UTC (rev 2)',
      '2011-04-29 02:57:23 UTC (rev 3)',
      'Modified: content/textfile.txt')
    chk = [
      (api.Event.PATCH_START, patch),
      (api.Event.ENTRY_START, entry),
      (api.Event.LINE_LOC, api.Line('@@ -1 +1 @@')),
      (api.Event.LINE_DELETE, api.Line('this is a sample textfile.txt.', oldnum=1, segments=[
         (api.Event.SEGMENT_SAME, 'this is a sample textfile.txt.'),
         ])),
      (api.Event.LINE_ADD, api.Line('this is a sample textfile.txt with changes.', newnum=1, segments=[
         (api.Event.SEGMENT_SAME, 'this is a sample textfile.txt'),
         (api.Event.SEGMENT_ADD, ' with changes'),
         (api.Event.SEGMENT_SAME, '.'),
         ])),
      (api.Event.ENTRY_END, entry),
      (api.Event.PATCH_END, patch),
      ]
    gen = parsedifflib.parse_svnlook(src, {'lineNumbers': True, 'intraLineDiff': True})
    chk = [(type, repr(obj)) for type, obj in chk]
    out = [(type, repr(obj)) for type, obj in gen]
    self.assertEqual(out, chk)

  #----------------------------------------------------------------------------
  def test_property_default(self):
    src = self.propEntry_diff
    patch = api.Patch()
    entry = api.Entry(
      api.Entry.TYPE_PROPERTY,
      None, None, None, None,
      'Property changes on: property/file1.txt')
    propent = api.PropertyEntry(
      'Modified: test:property1',
      old = 'one line\n\n(blank line above and below too)\n\n',
      new = 'one line\n(blank line above removed)\n(blank line below)\n\n')
    chk = [
      (api.Event.PATCH_START, patch),
      (api.Event.ENTRY_START, entry),
      (api.Event.PROPENTRY, propent),
      (api.Event.ENTRY_END, entry),
      (api.Event.PATCH_END, patch),
      ]
    gen = parsedifflib.parse_svnlook(src)
    chk = [(type, repr(obj)) for type, obj in chk]
    out = [(type, repr(obj)) for type, obj in gen]
    self.assertEqual(out, chk)

  #----------------------------------------------------------------------------
  def test_property_propertyUnifiedDiff(self):
    src = self.propEntry_diff
    patch = api.Patch()
    entry = api.Entry(
      api.Entry.TYPE_PROPERTY,
      None, None, None, None,
      'Property changes on: property/file1.txt')
    propent = api.PropertyEntry(
      'Modified: test:property1',
      old = 'one line\n\n(blank line above and below too)\n\n',
      new = 'one line\n(blank line above removed)\n(blank line below)\n\n')
    chk = [
      (api.Event.PATCH_START, patch),
      (api.Event.ENTRY_START, entry),
      (api.Event.PROPENTRY_START, propent),
      (api.Event.LINE_LOC, api.Line('@@ -1,5 +1,5 @@')),
      (api.Event.LINE_SAME, api.Line('one line')),
      (api.Event.LINE_DELETE, api.Line('')),
      (api.Event.LINE_DELETE, api.Line('(blank line above and below too)')),
      (api.Event.LINE_ADD, api.Line('(blank line above removed)')),
      (api.Event.LINE_ADD, api.Line('(blank line below)')),
      (api.Event.LINE_SAME, api.Line('')),
      (api.Event.LINE_SAME, api.Line('')),
      (api.Event.PROPENTRY_END, propent),
      (api.Event.ENTRY_END, entry),
      (api.Event.PATCH_END, patch),
      ]
    gen = parsedifflib.parse_svnlook(src, {'propertyUnifiedDiff': True})
    chk = [(type, repr(obj)) for type, obj in chk]
    out = [(type, repr(obj)) for type, obj in gen]
    self.assertEqual(out, chk)

  #----------------------------------------------------------------------------
  def test_property_propertyUnifiedDiff_lineNumbers(self):
    src = self.propEntry_diff
    patch = api.Patch()
    entry = api.Entry(
      api.Entry.TYPE_PROPERTY,
      None, None, None, None,
      'Property changes on: property/file1.txt')
    propent = api.PropertyEntry(
      'Modified: test:property1',
      old = 'one line\n\n(blank line above and below too)\n\n',
      new = 'one line\n(blank line above removed)\n(blank line below)\n\n')
    chk = [
      (api.Event.PATCH_START, patch),
      (api.Event.ENTRY_START, entry),
      (api.Event.PROPENTRY_START, propent),
      (api.Event.LINE_LOC, api.Line('@@ -1,5 +1,5 @@')),
      (api.Event.LINE_SAME, api.Line('one line', oldnum=1, newnum=1)),
      (api.Event.LINE_DELETE, api.Line('', oldnum=2)),
      (api.Event.LINE_DELETE, api.Line('(blank line above and below too)', oldnum=3)),
      (api.Event.LINE_ADD, api.Line('(blank line above removed)', newnum=2)),
      (api.Event.LINE_ADD, api.Line('(blank line below)', newnum=3)),
      (api.Event.LINE_SAME, api.Line('', oldnum=4, newnum=4)),
      (api.Event.LINE_SAME, api.Line('', oldnum=5, newnum=5)),
      (api.Event.PROPENTRY_END, propent),
      (api.Event.ENTRY_END, entry),
      (api.Event.PATCH_END, patch),
      ]
    gen = parsedifflib.parse_svnlook(src, {
      'lineNumbers': True, 'propertyUnifiedDiff': True})
    chk = [(type, repr(obj)) for type, obj in chk]
    out = [(type, repr(obj)) for type, obj in gen]
    self.assertEqual(out, chk)

  #----------------------------------------------------------------------------
  def runtest(self, item):
    out = colorize(asset.load('parsedifflib:' + item.name).read())
    chk = asset.load('parsedifflib:' + item.name + '.output.reference').read()
    self.assertMultiLineEqual(out, chk)

for item in asset.load('parsedifflib:data/*.diff'):
  safename = re.sub('[^a-zA-Z0-9]+', '_', item.name[5:-5])
  def make_test(item):
    def test_asset(self):
      self.runtest(item)
    return test_asset
  setattr(TestParseDiffLib, 'test_auto_' + safename, make_test(item))

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: metagriffin <mg.github@uberdev.org>
# date: 2013/12/10
# copy: (C) Copyright 2013-EOT metagriffin -- see LICENSE.txt
#------------------------------------------------------------------------------
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
class Event(object):

  PATCH_START           = 10
  PATCH_END             = 11
  ENTRY_START           = 20
  ENTRY_END             = 21
  LINE_LOC              = 30
  LINE_SAME             = 31
  LINE_ADD              = 32
  LINE_DELETE           = 33
  LINE_NOTE             = 34
  PROPENTRY             = 40
  PROPENTRY_START       = 41
  PROPENTRY_END         = 42
  SEGMENT_SAME          = 51
  SEGMENT_ADD           = 52
  SEGMENT_DELETE        = 53

#------------------------------------------------------------------------------
class SimpleRepr(object):
  def __repr__(self):
    ret = '<parsedifflib.%s' % (self.__class__.__name__,)
    for key in dir(self):
      if key.startswith('_'):
        continue
      try:
        val = getattr(self, key, None)
      except AttributeError:
        continue
      if val is not None:
        ret += ' %s=%r' % (key, val)
    return ret + '>'

#------------------------------------------------------------------------------
class Patch(SimpleRepr): pass

#------------------------------------------------------------------------------
class Entry(SimpleRepr):

  TYPE_CONTENT          = 1
  TYPE_PROPERTY         = 2

  def __init__(self, type,
               source=None, target=None,
               srcsig=None, tgtsig=None,
               comment=None):
    self.type    = type
    self.source  = source
    self.target  = target
    self.srcsig  = srcsig
    self.tgtsig  = tgtsig
    self.comment = comment

#------------------------------------------------------------------------------
class Line(SimpleRepr):
  def __init__(self, line, oldnum=None, newnum=None, segments=None):
    self.line     = line
    self.oldnum   = oldnum
    self.newnum   = newnum
    self.segments = segments

#------------------------------------------------------------------------------
class Segment(SimpleRepr):
  def __init__(self, text):
    self.text = text
  def __repr__(self):
    return repr(self.text)

#------------------------------------------------------------------------------
class PropertyEntry(SimpleRepr):
  def __init__(self, head, old=None, new=None):
    self.head = head
    self.old  = old
    self.new  = new

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
#------------------------------------------------------------------------------
# file: $Id$
# auth: metagriffin <mg.github@uberdev.org>
# date: 2013/11/27
# copy: (C) Copyright 2013-EOT metagriffin -- see LICENSE.txt
#------------------------------------------------------------------------------
# This software is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
#------------------------------------------------------------------------------

import six, morph, re, difflib
from aadict import aadict

from . import api
from .util import Source

#------------------------------------------------------------------------------
class ParseError(Exception):
  def __init__(self, msg, stage):
    super(ParseError, self).__init__(
      msg + ' on line %r: %s' % (stage.source.curline, stage.source.peekline()))
class InvalidChangeError(ParseError): pass

#------------------------------------------------------------------------------
def parse_svnlook(data, settings=None):
  stage = aadict(options=aadict(
    lineNumbers          = False,
    intraLineDiff        = False,
    intraLineCompact     = False,
    propertyUnifiedDiff  = False,
    ))
  stage.options.update(settings or {})
  if morph.isstr(data):
    data = six.StringIO(data)
  stage.source = Source(data)
  events = _parse(stage)
  if stage.options.propertyUnifiedDiff:
    events = _addPropertyUnifiedDiff(stage, events)
  if stage.options.lineNumbers:
    events = _addLineNumbers(stage, events)
  if stage.options.intraLineDiff:
    events = _addIntraLineDiff(stage, events)
  for evt in events:
    yield evt

#------------------------------------------------------------------------------
_addLineNumbers_loc_re = re.compile('^@@ -(\d+)(?:,\d+)? \+(\d+)(?:,\d+)? @@$')
def _addLineNumbers(stage, events):
  pos = [None, None]
  for etype, obj in events:
    if etype == api.Event.LINE_LOC:
      match = _addLineNumbers_loc_re.match(obj.line)
      if not match:
        raise ValueError('invalid location line: %r' % (obj.line,))
      pos = map(int, match.groups())
    if etype == api.Event.LINE_SAME:
      obj.oldnum = pos[0]
      obj.newnum = pos[1]
      pos[0] += 1
      pos[1] += 1
    if etype == api.Event.LINE_ADD:
      obj.newnum = pos[1]
      pos[1] += 1
    if etype == api.Event.LINE_DELETE:
      obj.oldnum = pos[0]
      pos[0] += 1
    yield etype, obj

#------------------------------------------------------------------------------
def _addIntraLineDiff(stage, events):
  buf = []
  def handleBuffer(buf):
    # todo: a *very* smart implementation would investigate consecutive
    # adds and deletes to determine if they should be interleaved...
    for idx, cur in enumerate(buf):
      if cur[0] == api.Event.LINE_DELETE \
          and ( idx == 0 or buf[idx - 1][0] != api.Event.LINE_DELETE ) \
          and ( idx + 1 < len(buf) ) and buf[idx + 1][0] == api.Event.LINE_ADD \
          and ( ( idx + 2 >= len(buf) ) or buf[idx + 2][0] != api.Event.LINE_ADD ):
        _calcIntraLineDiff(stage, cur[1], buf[idx + 1][1])
    buf.reverse()
    while buf:
      yield buf.pop()
  for typ, obj in events:
    if typ not in (api.Event.LINE_ADD, api.Event.LINE_DELETE):
      for evt in handleBuffer(buf):
        yield evt
      yield typ, obj
      continue
    buf.append((typ, obj))
    continue
  # paranoia. this should never be needed, since there should have
  # been, at the very least, a terminal PATCH_END event triggering a
  # handleBuffer() earlier.
  for evt in handleBuffer(buf):
    yield evt

#------------------------------------------------------------------------------
def _calcIntraLineDiff(stage, rem, add):
  seq = difflib.SequenceMatcher(a=rem.line, b=add.line)
  rems = []
  adds = []
  for op in seq.get_opcodes():
    if op[0] == 'equal':
      rems.append((api.Event.SEGMENT_SAME, api.Segment(rem.line[op[1]:op[2]])))
      adds.append((api.Event.SEGMENT_SAME, api.Segment(add.line[op[3]:op[4]])))
      continue
    if op[0] == 'insert':
      adds.append((api.Event.SEGMENT_ADD, api.Segment(add.line[op[3]:op[4]])))
      continue
    if op[0] == 'delete':
      rems.append((api.Event.SEGMENT_DELETE, api.Segment(rem.line[op[1]:op[2]])))
      continue
    if op[0] == 'replace':
      rems.append((api.Event.SEGMENT_DELETE, api.Segment(rem.line[op[1]:op[2]])))
      adds.append((api.Event.SEGMENT_ADD, api.Segment(add.line[op[3]:op[4]])))
      continue
    raise ValueError('unexpected difflib opcode %r' % (op[0],))
  rem.segments = reduce(_collapseSegments, rems, [])
  add.segments = reduce(_collapseSegments, adds, [])

#------------------------------------------------------------------------------
def _collapseSegments(ret, item):
  if not ret or ret[-1][0] != item[0]:
    return ret + [item]
  ret[-1][1].text += item[1].text
  return ret

#------------------------------------------------------------------------------
def _addPropertyUnifiedDiff(stage, events):
  for typ, prop in events:
    if typ != api.Event.PROPENTRY:
      yield typ, prop
      continue
    yield api.Event.PROPENTRY_START, prop
    old = prop.old.split('\n') if prop.old is not None else []
    new = prop.new.split('\n') if prop.new is not None else []
    for ops in difflib.SequenceMatcher(a=old, b=new).get_grouped_opcodes():
      yield api.Event.LINE_LOC, api.Line('@@ -%d,%d +%d,%d @@' % (
        ops[0][1] + 1, ops[-1][2] - ops[0][1],
        ops[0][3] + 1, ops[-1][4] - ops[0][3]))
      for op in ops:
        if op[0] == 'equal':
          for idx in range(op[1], op[2]):
            yield api.Event.LINE_SAME, api.Line(old[idx])
          continue
        if op[0] == 'insert':
          for idx in range(op[3], op[4]):
            yield api.Event.LINE_ADD, api.Line(new[idx])
          continue
        if op[0] == 'delete':
          for idx in range(op[1], op[2]):
            yield api.Event.LINE_DELETE, api.Line(old[idx])
          continue
        if op[0] == 'replace':
          for idx in range(op[1], op[2]):
            yield api.Event.LINE_DELETE, api.Line(old[idx])
          for idx in range(op[3], op[4]):
            yield api.Event.LINE_ADD, api.Line(new[idx])
          continue
        raise ValueError('unexpected difflib opcode %r' % (op[0],))
    yield api.Event.PROPENTRY_END, prop

#------------------------------------------------------------------------------
def _parse(stage):
  stage.patch = api.Patch()
  yield api.Event.PATCH_START, stage.patch
  while stage.source.peekline():
    for evt in _parse_change(stage):
      yield evt
  yield api.Event.PATCH_END, stage.patch

#------------------------------------------------------------------------------
def _parse_change(stage):
  found = False
  for evt in _parse_svnlook_contentChange(stage):
    found = True
    yield evt
  if found:
    return
  for evt in _parse_svnlook_propertyChange(stage):
    found = True
    yield evt
  if found:
    return
  raise InvalidChangeError('invalid source', stage)

#------------------------------------------------------------------------------
def _parse_svnlook_contentChange(stage):
  peek = stage.source.peeklines(2)
  if not peek[0].strip() or peek[1] != ( '=' * 67 + '\n' ):
    return
  comment, sep, old, new = stage.source.readlines(4)

  # check for empty file events
  if old == '\n' or ( not old and not new ):
    entry = api.Entry(api.Entry.TYPE_CONTENT, comment=comment[:-1])
    yield api.Event.ENTRY_START, entry
    yield api.Event.ENTRY_END, entry
    stage.source.pushlines([old, new])
    return

  # check for binary file events
  if old == '(Binary files differ)\n' and new == '\n':
    entry = api.Entry(api.Entry.TYPE_CONTENT, comment=comment[:-1])
    yield api.Event.ENTRY_START, entry
    yield api.Event.LINE_NOTE, api.Line(old[:-1])
    yield api.Event.ENTRY_END, entry
    return

  # if not empty or binary, ensure proper begin declaration
  if not ( old.startswith('--- ') and new.startswith('+++ ') ):
    stage.source.pushlines([comment, sep, old, new])
    return

  # and now parse the changes
  entry = api.Entry(
    api.Entry.TYPE_CONTENT,
    source = old[4:].split('\t', 1)[0],
    target = new[4:].split('\t', 1)[0],
    srcsig = old[4:].split('\t', 1)[1][:-1],
    tgtsig = new[4:].split('\t', 1)[1][:-1],
    comment = comment[:-1])
  yield api.Event.ENTRY_START, entry

  while True:
    line = stage.source.readline()
    if not line or line == '\n':
      break
    if line.startswith('@@ '):
      yield api.Event.LINE_LOC, api.Line(line.strip())
      continue
    if line.startswith(' '):
      yield api.Event.LINE_SAME, api.Line(line[1:-1])
      continue
    if line.startswith('\\') or line.startswith('/'):
      yield api.Event.LINE_NOTE, api.Line(line[:-1])
      continue
    if line.startswith('+'):
      yield api.Event.LINE_ADD, api.Line(line[1:-1])
      continue
    if line.startswith('-'):
      yield api.Event.LINE_DELETE, api.Line(line[1:-1])
      continue
    raise ParseError('unexpected prefix', stage)

  yield api.Event.ENTRY_END, entry

propChangeEntry_re = re.compile('^(Added|Modified|Deleted): (.+)$')

#------------------------------------------------------------------------------
def _parse_svnlook_propertyChange(stage):
  peek = stage.source.peeklines(3)
  if peek[0] != '\n' or not peek[1].startswith('Property changes on: ') \
      or peek[2] != ( '_' * 67 + '\n' ):
    return
  nl, comment, sep = stage.source.readlines(3)

  entry = api.Entry(api.Entry.TYPE_PROPERTY, comment=comment[:-1])
  yield api.Event.ENTRY_START, entry

  prop = api.PropertyEntry(None)

  while True:
    line = stage.source.readline()
    if propChangeEntry_re.match(line):
      if prop.head:
        if prop.old is None and prop.new is None:
          raise ParseError('unexpected property entry (head)', stage)
        yield api.Event.PROPENTRY, prop
      prop.head, prop.old, prop.new = line[:-1], None, None
      continue
    if not prop.head:
      raise ParseError('unexpected property entry (no-head)', stage)
    if line.startswith('   - '):
      if prop.old:
        raise ParseError('unexpected property entry (repeat-old)', stage)
      if prop.new:
        raise ParseError('unexpected property entry (new-first)', stage)
      prop.old = line[5:-1]
      continue
    if line.startswith('   + '):
      if prop.new:
        raise ParseError('unexpected property entry (repeat-new)', stage)
      prop.new = line[5:-1]
      continue
    if line == '\n':
      if not stage.source.peekline():
        if prop.old is None and prop.new is None:
          raise ParseError('unexpected EOF in property entry', stage)
        yield api.Event.PROPENTRY, prop
        yield api.Event.ENTRY_END, entry
        return
      try:
        for evt in _parse_change(stage):
          if prop is not None:
            if prop.old is None and prop.new is None:
              raise ParseError('unexpected property entry (empty)', stage)
            yield api.Event.PROPENTRY, prop
            yield api.Event.ENTRY_END, entry
            prop = None
          yield evt
        return
      except InvalidChangeError:
        pass
    if prop.new is not None:
      prop.new += '\n' + line[:-1]
      continue
    if prop.old is not None:
      prop.old += '\n' + line[:-1]
      continue
    raise ParseError('unexpected property entry (no-new-or-old)', stage)

  raise ParseError('unexpected property entry (no-term)', stage)

#------------------------------------------------------------------------------
# end of $Id$
#------------------------------------------------------------------------------
